"""
========================================================================================
Wrap landlab component with the Basic Modeling Interface (:mod:`landlab.bmi.bmi_bridge`)
========================================================================================

.. sectionauthor:: Eric Hutton

Function reference
------------------

The `wrap_as_bmi` function wraps a landlab component class so that it
exposes a Basic Modelling Interface.

"""
import os

import numpy as np
import yaml

from ..core.model_component import Component
from ..grid import RasterModelGrid

__all__ = ['TimeStepper', 'wrap_as_bmi']


class TimeStepper(object):

    """Step through time.

    Parameters
    ----------
    start : float, optional
        Clock start time.
    stop : float, optional
        Stop time.
    step : float, optional
        Time step.

    Examples
    --------
    >>> from landlab.bmi import TimeStepper
    >>> time_stepper = TimeStepper()
    >>> time_stepper.start
    0.0
    >>> time_stepper.stop is None
    True
    >>> time_stepper.step
    1.0
    >>> time_stepper.time
    0.0
    >>> for _ in range(10): time_stepper.advance()
    >>> time_stepper.time
    10.0
    >>> time_stepper = TimeStepper(1., 13., 2.)
    >>> [time for time in time_stepper]
    [1.0, 3.0, 5.0, 7.0, 9.0, 11.0]
    """

    def __init__(self, start=0., stop=None, step=1.):
        self._start = start
        self._stop = stop
        self._step = step

        self._time = start

    def __iter__(self):
        if self.stop is None:
            while 1:
                yield self._time
                self._time += self._step
        else:
            while self._time < self._stop:
                yield self._time
                self._time += self._step
        raise StopIteration()

    @property
    def time(self):
        """Current time."""
        return self._time

    @property
    def start(self):
        """Start time."""
        return self._start

    @property
    def stop(self):
        """Stop time."""
        return self._stop

    @property
    def step(self):
        """Time Step."""
        return self._step

    @step.setter
    def step(self, new_val):
        """Change the time step."""
        self._step = new_val

    def advance(self):
        """Advance the time stepper by one time step."""
        self._time += self.step
        if self._stop is not None and self._time > self._stop:
            raise StopIteration()


def wrap_as_bmi(cls):
    """Wrap a landlab class so it exposes a BMI.

    Parameters
    ----------
    cls : class
        A landlab class that inherits from `Component`.

    Returns
    -------
    class
        A wrapped class that exposes a BMI.

    Examples
    --------
    >>> from landlab.bmi import wrap_as_bmi
    >>> from landlab.components.flexure import Flexure

    >>> BmiFlexure = wrap_as_bmi(Flexure)
    >>> flexure = BmiFlexure()

    >>> config = \"\"\"
    ... eet: 10e3
    ... method: flexure
    ... clock:
    ...     start: 0.
    ...     stop: 10.
    ...     step: 2.
    ... grid:
    ...     type: raster
    ...     shape: [20, 40]
    ...     spacing: [1000., 2000.]
    ... \"\"\"
    >>> flexure.initialize(config)
    >>> flexure.get_output_var_names()
    ('lithosphere_surface__elevation_increment',)
    >>> flexure.get_var_grid('lithosphere_surface__elevation_increment')
    0
    >>> flexure.get_grid_shape(0)
    (20, 40)
    >>> dz = flexure.get_value('lithosphere_surface__elevation_increment')
    >>> dz.shape == (800, )
    True

    >>> np.all(dz == 0.)
    True
    >>> flexure.get_current_time()
    0.0

    >>> flexure.get_input_var_names()
    ('lithosphere__overlying_pressure_increment',)
    >>> load = np.zeros((20, 40), dtype=float)
    >>> load[0, 0] = 1.
    >>> flexure.set_value('lithosphere__overlying_pressure_increment', load)
    >>> flexure.update()
    >>> flexure.get_current_time()
    2.0
    >>> dz = flexure.get_value('lithosphere_surface__elevation_increment')
    >>> np.all(dz == 0.)
    False
    """
    if not issubclass(cls, Component):
        raise TypeError('class must inherit from Component')

    class BmiWrapper(object):
        __doc__ = """
        Basic Modeling Interface for the {name} component.
        """.format(name=cls.__name__).strip()

        _cls = cls
        def __init__(self):
            self._base = None
            self._clock = None
            super(BmiWrapper, self).__init__()

        def get_component_name(self):
            """Name of the component."""
            return self._cls.name

        def get_input_var_names(self):
            """Names of the input exchange items."""
            return self._cls.input_var_names

        def get_output_var_names(self):
            """Names of the output exchange items."""
            return self._cls.output_var_names

        def get_current_time(self):
            """Current component time."""
            return self._clock.time

        def get_end_time(self):
            """Stop time for the component."""
            return self._clock.stop

        def get_start_time(self):
            """Start time of the component."""
            return self._clock.start

        def get_time_step(self):
            """Component time step."""
            return self._clock.step

        def get_time_units(self):
            """Time units used by the component."""
            raise NotImplementedError('get_time_units not implemented')

        def initialize(self, fname):
            """Initialize the component from a file.

            BMI-wrapped Landlab components use input files in YAML format.
            Component-specific parameters are listed at the top level,
            followed by grid and then time information. An example input
            file looks like::

                eet: 15e3
                clock:
                    start: 0
                    stop: 100.
                    step: 2.
                grid:
                    type: raster
                    shape: [20, 40]
                    spacing: [1000., 2000.]

            In this case, a `RasterModelGrid` is created (with the given shape
            and spacing) and passed to the underlying landlab component. The
            `eet=15000.` is also given to the component but as a keyword
            parameter. The BMI clock is initialized with the given parameters.

            Parameters
            ----------
            fname : str or file_like
                YAML-formatted input file for the component.
            """
            if os.path.isfile(fname):
                with open(fname, 'r') as fp:
                    params = yaml.load(fp)
            else:
                params = yaml.load(fname)

            grid_params = params.pop('grid')
            gtype = grid_params.pop('type')
            if gtype == 'raster':
                cls = RasterModelGrid
            else:
                raise ValueError(
                    'unrecognized grid type {gtype}'.format(gtype=gtype))

            grid = cls.from_dict(grid_params)

            clock_params = params.pop('clock')
            self._clock = TimeStepper(**clock_params)

            self._base = self._cls(grid, **params)

        def update(self):
            """Update the component one time step."""
            if hasattr(self._base, 'update'):
                self._base.update()
            self._clock.advance()

        def update_frac(self, frac):
            """Update the component a fraction of a time step."""
            time_step = self.get_time_step()
            self._clock.step = time_step * frac
            self.update()
            self._clock.step = time_step

        def update_until(self, then):
            """Update the component until a given time."""
            n_steps = (then - self.get_current_time()) / self.get_time_step()
            for _ in xrange(int(n_steps)):
                self.update()
            self.update_frac(n_steps - int(n_steps))

        def finalize(self):
            """Clean-up the component."""
            pass

        def get_var_grid(self, name):
            """Get the grid id for a variable."""
            return 0

        def get_var_itemsize(self, name):
            """Get the size of elements of a variable."""
            return np.dtype('float').itemsize

        def get_var_nbytes(self, name):
            """Get the total number of bytes used by a variable."""
            return self.get_itemsize(name) * self._base.grid.number_of_nodes

        def get_var_type(self, name):
            """Get the data type for a variable."""
            return str(np.dtype('float'))

        def get_var_units(self, name):
            """Get the unit used by a variable."""
            return self._cls.var_units(name)

        def get_value_ref(self, name):
            """Get a reference to a variable's data."""
            return self._base.grid.at_node[name]

        def get_value(self, name):
            """Get a copy of a variable's data."""
            return self._base.grid.at_node[name].copy()

        def set_value(self, name, vals):
            """Set the values of a variable."""
            if name in self.get_input_var_names():
                if name in self._base.grid.at_node:
                    self._base.grid.at_node[name][:] = vals.flat
                else:
                    self._base.grid.at_node[name] = vals
            else:
                raise KeyError('{name} is not an input item'.format(name=name))

        def get_grid_origin(self, gid):
            """Get the origin for a structured grid."""
            return (self._base.grid.node_y[0], self._base.grid.node_x[0])

        def get_grid_rank(self, gid):
            """Get the number of dimensions of a grid."""
            return 2

        def get_grid_shape(self, gid):
            """Get the shape of a structured grid."""
            return (self._base.grid.number_of_node_rows,
                    self._base.grid.number_of_node_columns)

        def get_grid_spacing(self, gid):
            """Get the row and column spacing of a structured grid."""
            return (self._base.grid.dy, self._base.grid.dx)

        def get_grid_type(self, gid):
            """Get the type of grid."""
            return 'uniform_rectilinear'


    BmiWrapper.__name__ = cls.__name__
    return BmiWrapper



#! /usr/bin/env python
"""
Unit tests for landlab.components.flexure.flexure
"""
from nose.tools import assert_equal, assert_true, assert_raises, with_setup
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import numpy as np

from landlab import RasterModelGrid
from landlab.components.flexure.flexure import Flexure


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e3, 10e3), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def setup_grid():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((20, 20), spacing=10e3)
    flex = Flexure(grid)
    globals().update({
        'flex': Flexure(grid)
    })


@with_setup(setup_grid)
def test_name():
    assert_equal(flex.name, 'Flexure')


@with_setup(setup_grid)
def test_input_var_names():
    assert_equal(flex.input_var_names,
                 ('lithosphere__overlying_pressure_increment', ))


@with_setup(setup_grid)
def test_output_var_names():
    assert_equal(flex.output_var_names,
                 ('lithosphere_surface__elevation_increment',))


@with_setup(setup_grid)
def test_var_units():
    assert_equal(set(flex.input_var_names) |
                 set(flex.output_var_names),
                 set(dict(flex.units).keys()))

    assert_equal(flex.var_units('lithosphere_surface__elevation_increment'), 'm')
    assert_equal(flex.var_units('lithosphere__overlying_pressure_increment'),
                 'Pa')


@with_setup(setup_grid)
def test_grid_shape():
    assert_equal(flex.grid.number_of_node_rows, _SHAPE[0])
    assert_equal(flex.grid.number_of_node_columns, _SHAPE[1])


@with_setup(setup_grid)
def test_grid_x_extent():
    assert_equal(flex.grid.extent[1], (_SHAPE[1] - 1) * _SPACING[1])


@with_setup(setup_grid)
def test_grid_y_extent():
    assert_equal(flex.grid.extent[0], (_SHAPE[0] - 1) * _SPACING[0])


@with_setup(setup_grid)
def test_field_getters():
    for name in flex.grid['node']:
        field = flex.grid['node'][name]
        assert_is_instance(field, np.ndarray)
        assert_equal(field.shape,
                     (flex.grid.number_of_node_rows *
                      flex.grid.number_of_node_columns, ))

    assert_raises(KeyError, lambda: flex.grid['not_a_var_name'])


@with_setup(setup_grid)
def test_field_initialized_to_zero():
    for name in flex.grid['node']:
        field = flex.grid['node'][name]
        assert_true(np.all(field == 0.))



#! /usr/bin/env python
"""Calculate slope aspects on a :any:`RasterModelGrid`."""
import numpy as np
from landlab.utils.decorators import deprecated


def _one_line_slopes(input_array, grid, vals):
    node = input_array[0]
    diagonals = input_array[5:]
    neighbors = input_array[1:5]

    if not grid.status_at_node[node] == 0:
        raise IndexError('One or more of the provided nodes was closed!')

    try:
        slope_we = (
            (vals[diagonals[1]] + 2. * vals[neighbors[2]] +
             vals[diagonals[2]]) - (vals[diagonals[0]] +
                                    2. * vals[neighbors[0]] +
                                    vals[diagonals[3]])
        ) / (8. * grid.dx)
        slope_sn = (
            (vals[diagonals[2]] + 2. * vals[neighbors[3]] +
             vals[diagonals[3]]) - (vals[diagonals[1]] +
                                    2. * vals[neighbors[:, 1]] +
                                    vals[diagonals[0]])
        ) / (8. * grid.dy)
        return slope_we, slope_sn
    except IndexError:
        C = vals[node]
        weighting_verticals = 4.
        weighting_horizontals = 4.
        try:
            vertical_grad = (vals[neighbors[3]] -
                             vals[neighbors[1]]) / (2. * grid.dy)
        except IndexError:
            try:
                vertical_grad = (C - vals[neighbors[1]]) / grid.dy
            except IndexError:
                try:
                    vertical_grad = (vals[neighbors[3]] - C) / grid.dy
                except IndexError:
                    vertical_grad = 0.
                    weighting_verticals -= 2.
        try:
            horizontal_grad = (
                (vals[neighbors[2]] - vals[neighbors[0]]) /
                (2. * grid.dx))
        except IndexError:
            try:
                horizontal_grad = (C - vals[neighbors[0]]) / grid.dx
            except IndexError:
                try:
                    horizontal_grad = (
                        vals[neighbors[2]] - C) / grid.dx
                except IndexError:
                    horizontal_grad = 0.
                    weighting_horizontals -= 2.
        try:
            left_grad = (vals[diagonals[2]] -
                         vals[diagonals[1]]) / (2. * grid.dx)
        except IndexError:
            try:
                C = vals[neighbors[2]]
            except:
                left_grad = 0.
                weighting_verticals -= 1.
            else:
                try:
                    left_grad = (C - vals[diagonals[1]]) / grid.dx
                except IndexError:
                    left_grad = (vals[diagonals[2]] - C) / grid.dx
        try:
            right_grad = (vals[diagonals[3]] -
                          vals[diagonals[0]]) / (2. * grid.dx)
        except IndexError:
            try:
                C = vals[neighbors[0]]
            except:
                right_grad = 0.
                weighting_verticals -= 1.
            else:
                try:
                    right_grad = (C - vals[diagonals[0]]) / grid.dx
                except IndexError:
                    right_grad = (vals[diagonals[3]] - C) / grid.dx
        try:
            top_grad = (vals[diagonals[1]] -
                        vals[diagonals[0]]) / (2. * grid.dy)
        except IndexError:
            try:
                C = vals[neighbors[1]]
            except:
                top_grad = 0.
                weighting_horizontals -= 1.
            else:
                try:
                    top_grad = (C - vals[diagonals[0]]) / grid.dy
                except IndexError:
                    top_grad = (vals[diagonals[1]] - C) / grid.dy
        try:
            bottom_grad = (vals[diagonals[2]] -
                           vals[diagonals[3]]) / (2. * grid.dy)
        except IndexError:
            try:
                C = vals[neighbors[3]]
            except:
                bottom_grad = 0.
                weighting_horizontals -= 1.
            else:
                try:
                    bottom_grad = (C - vals[diagonals[3]]) / grid.dy
                except IndexError:
                    bottom_grad = (vals[diagonals[2]] - C) / grid.dy

        slope_we = (top_grad + 2. * horizontal_grad +
                    bottom_grad) / weighting_horizontals
        slope_sn = (left_grad + 2. * vertical_grad +
                    right_grad) / weighting_verticals

        return slope_we, slope_sn


@deprecated(use='grid.calc_slope_at_node', version=1.0)
def calc_slope_aspect_of_nodes_horn(grid, ids=None,
                                    vals='topographic__elevation'):
    r"""Calculate slope and aspect.

    .. note::

        THIS CODE HAS ISSUES (SN 25-Sept-14): This code didn't perform well
        on a NS facing elevation profile. Please check
        slope_aspect_routines_comparison.py under landlab\examples before
        using this.  Suggested alternative:
        calculate_slope_aspect_at_nodes_burrough

    Calculates the local topographic slope (i.e., the down-dip slope, and
    presented as positive), and the aspect (dip direction in radians
    clockwise from north), at the given nodes, *ids*. All *ids* must be of
    core nodes.

    This method uses the Horn 1981 algorithm, the one employed by many GIS
    packages. It should be significantly faster than alternative slope
    methods.

    If *ids* is not provided, the slope will be returned for all core
    nodes.

    *vals* is either the name of an existing grid field from which to draw
    topographic data, or an array of values to use. If an array of values
    is passed, it must be nnodes long. If *vals* is not provided, this
    method will default to trying to use the field
    "topographic__elevation".

    Parameters
    ----------
    grid : RasterModelGrid
        Grid on which to calculate slopes and aspects.
    ids : array_like of int, optional
        Nodes on which to calculate slope and aspect.
    vals : str or ndarray, optional
        Node values used to measure slope and aspect.

    Returns
    -------
    (slope, aspect) : tuple of float
        *slope*: a len(ids) array of slopes at each node provided.
        *aspect*: a len(ids) array of aspects at each node provided.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> grid = RasterModelGrid((4, 5))

    Create a south-facing slope.

    >>> elevation = np.array([
    ...     0., 0., 0., 0., 0,
    ...     1., 1., 1., 1., 1,
    ...     2., 2., 2., 2., 2,
    ...     3., 3., 3., 3., 3])
    >>> (slope, aspect) = calc_slope_aspect_of_nodes_horn(grid,
    ...     vals=elevation)
    >>> len(slope) == grid.number_of_core_nodes
    True
    >>> len(aspect) == grid.number_of_core_nodes
    True
    >>> slope
    array([ 1.,  1.,  1.,  1.,  1.,  1.])
    >>> aspect * 180. / np.pi
    array([ 180.,  180.,  180.,  180.,  180.,  180.])

    Make the slope north-facing by multiplying elevations by -1. The slopes
    will still be positive but the aspects will change.

    >>> elevation *= -1
    >>> (slope, aspect) = calc_slope_aspect_of_nodes_horn(grid,
    ...     vals=elevation)
    >>> slope
    array([ 1.,  1.,  1.,  1.,  1.,  1.])
    >>> aspect * 180. / np.pi
    array([ 0.,  0.,  0.,  0.,  0.,  0.])

    Double the slope and make it west-facing.

    >>> elevation = np.array([
    ...     0., 1., 2., 3., 4.,
    ...     0., 1., 2., 3., 4.,
    ...     0., 1., 2., 3., 4.,
    ...     0., 1., 2., 3., 4.])
    >>> elevation *= 2.
    >>> (slope, aspect) = calc_slope_aspect_of_nodes_horn(grid,
    ...     vals=elevation)
    >>> slope
    array([ 2.,  2.,  2.,  2.,  2.,  2.])
    >>> aspect * 180. / np.pi
    array([ 270.,  270.,  270.,  270.,  270.,  270.])

    Make the slope east-facing by multiplying elevations by -1. The slopes
    will still be positive but the aspects will change.

    >>> elevation *= -1.
    >>> (slope, aspect) = calc_slope_aspect_of_nodes_horn(grid,
    ...     vals=elevation)
    >>> slope
    array([ 2.,  2.,  2.,  2.,  2.,  2.])
    >>> aspect * 180. / np.pi
    array([ 90.,  90.,  90.,  90.,  90.,  90.])
    """
    if ids is None:
        ids = grid.core_nodes
    if isinstance(vals, str):
        vals = grid.at_node[vals]
    else:
        if len(vals) != grid.number_of_nodes:
            raise IndexError('*vals* was not of a compatible length!')

    # [right, top, left, bottom]
    neighbors = grid.active_neighbors_at_node(ids)
    # [topright, topleft, bottomleft, bottomright]
    diagonals = grid._get_diagonal_list(ids)

    input_array = np.empty((len(ids), 9), dtype=int)
    input_array[:, 0] = ids
    input_array[:, 1:5] = neighbors
    input_array[:, 5:] = diagonals

    slopes_array = np.apply_along_axis(_one_line_slopes, 1, input_array, grid,
                                       vals)
    slope_we = slopes_array[:, 0]
    slope_sn = slopes_array[:, 1]

    slope = np.sqrt(slope_we * slope_we + slope_sn * slope_sn)
    # aspect = np.empty_like(slope)
    aspect = np.ones(slope.size, dtype=float)
    simple_cases = slope_we != 0.
    complex_cases = np.logical_not(simple_cases)

    complex_aspects = aspect[complex_cases]
    complex_aspects[slope_sn[complex_cases] < 0.] = np.pi
    complex_aspects[slope_sn[complex_cases] >= 0.] = 0.
    aspect[complex_cases] = complex_aspects

    # +ve is CCW rotation from x axis
    angle_to_xaxis = np.arctan(slope_sn[simple_cases] / slope_we[simple_cases])
    aspect[simple_cases] = (
        (1. - np.sign(slope_we[simple_cases])) * 0.5) * np.pi + (
            0.5 * np.pi - angle_to_xaxis)

    return slope.ravel(), aspect.ravel()



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import with_setup


def setup_grid():
    from landlab import RasterModelGrid
    globals().update({
        'rmg': RasterModelGrid(3, 3)
    })


@with_setup(setup_grid)
def test_lower_left_cell():
    assert_array_equal(rmg.nodes_around_point(.1, .1),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_upper_left_cell():
    assert_array_equal(rmg.nodes_around_point(.1, .9),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_upper_right_cell():
    assert_array_equal(rmg.nodes_around_point(.9, .9),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_lower_right_cell():
    assert_array_equal(rmg.nodes_around_point(.9, .1),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_on_left_edge():
    assert_array_equal(rmg.nodes_around_point(0., .5),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_on_bottom_edge():
    assert_array_equal(rmg.nodes_around_point(.5, 0.),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_on_right_edge():
    assert_array_equal(rmg.nodes_around_point(1., .5),
                       np.array([1, 4, 5, 2]))


@with_setup(setup_grid)
def test_on_top_edge():
    assert_array_equal(rmg.nodes_around_point(.5, 1.),
                       np.array([3, 6, 7, 4]))


@with_setup(setup_grid)
def test_on_lower_left_corner():
    assert_array_equal(rmg.nodes_around_point(0., 0.),
                       np.array([0, 3, 4, 1]))


@with_setup(setup_grid)
def test_on_upper_right_corner():
    assert_array_equal(rmg.nodes_around_point(1., 1.),
                       np.array([4, 7, 8, 5]))


@with_setup(setup_grid)
def test_on_upper_left_corner():
    assert_array_equal(rmg.nodes_around_point(0., 1.),
                       np.array([3, 6, 7, 4]))


@with_setup(setup_grid)
def test_on_lower_right_corner():
    assert_array_equal(rmg.nodes_around_point(1., 0.),
                       np.array([1, 4, 5, 2]))



import numpy as np
import six
from six.moves import range

from ..core.utils import make_optional_arg_into_id_array
from .structured_quad import links as squad_links


def neighbor_active_link_at_cell(grid, inds, *args):
    """neighbor_active_link_at_cell(grid, link_ids [, cell_ids])

    Return an array of the active link ids for neighbors of *cell_id* cells.
    *link_ids* is an index into the links of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.

    Parameters
    ----------
    grid : RasterModelGrid
        Source grid.
    link_inds : array_like
        IDs of links
    cell_ids : array_like, optional
        IDs of cells for which to get links

    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]
    links = grid._active_links_at_node(node_ids).T

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return links[range(len(cell_ids)), inds]


def neighbor_node_at_cell(grid, inds, *args):
    """ node_id_of_cell_neighbor(grid, neighbor_ids [, cell_ids])

    Return an array of the node ids for neighbors of *cell_id* cells.
    *neighbor_ids* is an index into the neighbors of a cell as measured
    clockwise starting from the south.

    If *cell_ids* is not given, return neighbors for all cells in the grid.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    neighbor_ids : array_like
        IDs of the neighbor nodes.
    cell_ids : array_like, optional
        IDs of cell about which to get neighbors.

    Returns
    -------
    ndarray
        Node IDs for given neighbors of cells.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_funcs import neighbor_node_at_cell
    >>> grid = RasterModelGrid(4, 5, 1.0)
    >>> neighbor_node_at_cell(grid, 0, 0)
    array([1])

    Get the lower and the the upper neighbors for all the cells.

    >>> neighbor_node_at_cell(grid, 0)
    array([1, 2, 3, 6, 7, 8])
    >>> neighbor_node_at_cell(grid, 2)
    array([11, 12, 13, 16, 17, 18])

    As an alternative to the above, use fancy-indexing to get both sets of
    neighbors with one call.

    >>> neighbor_node_at_cell(grid, np.array([0, 2]), [1, 4])
    array([[ 2, 12],
           [ 7, 17]])
    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]
    neighbors = grid.active_neighbors_at_node(node_ids)

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    # return neighbors[range(len(cell_ids)), 3 - inds]
    return (
        np.take(np.take(neighbors, range(len(cell_ids)), axis=0),
                3 - inds, axis=1))


def corner_node_at_cell(grid, inds, *args):
    """node_id_of_cell_corner(grid, corner_ids [, cell_ids])

    Return an array of the node ids for diagonal neighbors of *cell_id* cells.
    *corner_ids* is an index into the corners of a cell as measured
    clockwise starting from the southeast.

    If *cell_ids* is not given, return neighbors for all cells in the grid.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    corner_ids : array_like
        IDs of the corner nodes.
    cell_ids : array_like, optional
        IDs of cell about which to get corners

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_funcs import corner_node_at_cell
    >>> grid = RasterModelGrid(4, 5, 1.0)
    >>> corner_node_at_cell(grid, 0, 0)
    array([2])

    Get the lower-right and the the upper-left corners for all the cells.

    >>> corner_node_at_cell(grid, 0)
    array([2, 3, 4, 7, 8, 9])
    >>> corner_node_at_cell(grid, 2)
    array([10, 11, 12, 15, 16, 17])

    As an alternative to the above, use fancy-indexing to get both sets of
    corners with one call.

    >>> corner_node_at_cell(grid, np.array([0, 2]), [1, 4])
    array([[ 3, 11],
           [ 8, 16]])
    """
    cell_ids = make_optional_arg_into_id_array(grid.number_of_cells, *args)
    node_ids = grid.node_at_cell[cell_ids]
    diagonals = grid._get_diagonal_list(node_ids)

    if not isinstance(inds, np.ndarray):
        inds = np.array(inds)

    return (
        np.take(np.take(diagonals, range(len(cell_ids)), axis=0),
                3 - inds, axis=1))
    # return diagonals[range(len(cell_ids)), 3 - inds]


def calculate_flux_divergence_at_nodes(grid, active_link_flux, out=None):
    """Net flux into or out of nodes.

    Same as calculate_flux_divergence_at_core_cells, but works with and
    returns a list of net unit fluxes that corresponds to all nodes, rather
    than just core nodes.

    Parameters
    ----------
    grid : RasterModelGrid
        Input grid.
    active_link_flux : array_like
        Flux values at links.
    out : ndarray, optional
        Alternative output array in which to place the result.  Must
        be of the same shape and buffer length as the expected output.

    See Also
    --------
    calculate_flux_divergence_at_active_cells

    Notes
    -----
    Note that we DO compute net unit fluxes at boundary nodes (even though
    these don't have active cells associated with them, and often don't have
    cells of any kind, because they are on the perimeter). It's up to the
    user to decide what to do with these boundary values.

    Examples
    --------
    Calculate the gradient of values at a grid's nodes.

    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid((4, 5), spacing=1.0)
    >>> u = np.array([0., 1., 2., 3., 0.,
    ...               1., 2., 3., 2., 3.,
    ...               0., 1., 2., 1., 2.,
    ...               0., 0., 2., 2., 0.])
    >>> grad = rmg.calc_grad_at_link(u)[rmg.active_links]
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 1.,  1., -1.,
            1.,  1., -1.,  1.,
           -1., -1., -1.,
            1.,  1., -1.,  1.,
           -1.,  0.,  1.])

    Calculate the divergence of the gradients at each node.

    >>> flux = - grad    # downhill flux proportional to gradient
    >>> rmg.calculate_flux_divergence_at_nodes(flux)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 0., -1., -1.,  1.,  0.,
           -1.,  2.,  4., -2.,  1.,
           -1.,  0.,  1., -4.,  1.,
            0., -1.,  0.,  1.,  0.])

    If calculate_gradients_at_nodes is called inside a loop, you can
    improve speed by creating an array outside the loop. For example, do
    this once, before the loop:

    >>> df = rmg.zeros(centering='node') # outside loop
    >>> rmg.number_of_nodes
    20

    Then do this inside the loop so that the function will not have to create
    the df array but instead puts values into the *df* array.

    >>> df = rmg.calculate_flux_divergence_at_nodes(flux, out=df)

    >>> grid = RasterModelGrid((4, 5), spacing=(1, 2))
    >>> u = np.array([0., 1., 2., 3., 0.,
    ...               1., 2., 3., 2., 3.,
    ...               0., 1., 2., 1., 2.,
    ...               0., 0., 2., 2., 0.])
    >>> grad = grid.calc_grad_at_link(2 * u)[grid.active_links]
    >>> grad # doctest: +NORMALIZE_WHITESPACE
    array([ 2.,  2., -2.,
            1.,  1., -1.,  1.,
           -2., -2., -2.,
            1.,  1., -1., 1.,
           -2.,  0.,  2.])
    >>> grid.calculate_flux_divergence_at_nodes(- grad)
    ...     # doctest: +NORMALIZE_WHITESPACE
    array([ 0., -1., -1.,  1.,  0.,
           -1.,  2.,  4., -2.,  1.,
           -1.,  0.,  1., -4.,  1.,
            0., -1.,  0.,  1.,  0.])
    """
    assert len(active_link_flux) == grid.number_of_active_links, \
        "incorrect length of active_link_flux array"

    # If needed, create net_unit_flux array
    if out is None:
        out = grid.empty(centering='node')
    out.fill(0.)
    net_unit_flux = out

    assert len(net_unit_flux) == grid.number_of_nodes

    is_vert_link = squad_links.is_vertical_link(grid.shape, grid.active_links)
    vert_links = grid.active_links[is_vert_link]
    horiz_links = grid.active_links[~ is_vert_link]

    flux = np.zeros(grid.number_of_links + 1)

    flux[vert_links] = active_link_flux[is_vert_link] * grid.dy
    flux[horiz_links] = active_link_flux[~ is_vert_link] * grid.dx

    net_unit_flux[:] = (
        (flux[grid._node_active_outlink_matrix2[0][:]] +
         flux[grid._node_active_outlink_matrix2[1][:]]) -
        (flux[grid._node_active_inlink_matrix2[0][:]] +
         flux[grid._node_active_inlink_matrix2[1][:]])) / grid.cellarea

    return net_unit_flux


def calculate_slope_aspect_bfp(xs, ys, zs):
    """Calculate slope and aspect.

    .. codeauthor:: Katy Barnhart <katherine.barnhart@colorado.edu>

    Fits a plane to the given N points with given *xs*, *ys*, and *zs* values
    using single value decomposition.

    Returns a tuple of (*slope*, *aspect*) based on the normal vector to the
    best fit plane.

    .. note::

        This function does not check if the points fall on a line, rather
        than a plane.
    """
    if not len(xs) == len(ys) == len(zs):
        raise ValueError('array must be the same length')

    # step 1: subtract the centroid from the points
    # step 2: create a 3XN matrix of the points for SVD
    # in python, the unit normal to the best fit plane is
    # given by the third column of the U matrix.
    mat = np.vstack((xs - np.mean(xs), ys - np.mean(ys), zs - np.mean(zs)))
    U, _, _ = np.linalg.svd(mat)
    normal = U[:, 2]

    # step 3: calculate the aspect
    asp = 90.0 - np.degrees(np.arctan2(normal[1], normal[0]))
    asp = asp % 360.0

    # step 4: calculate the slope
    slp = 90.0 - np.degrees(np.arcsin(normal[2]))

    return slp, asp


def find_nearest_node(rmg, coords, mode='raise'):
    """Find the node nearest a point.

    Find the index to the node nearest the given x, y coordinates.
    Coordinates are provided as numpy arrays in the *coords* tuple.
    *coords* is tuple of coordinates, one for each dimension.

    The *mode* keyword to indicate what to do if a point is outside of the
    grid. Valid choices are the same as that used with the numpy function
    `ravel_multi_index`.

    A point is considered to be outside of the grid if it is outside the
    perimeter of the grid by one half of the grid spacing.

    Parameters
    ----------
    rmg : RasterModelGrid
        The source grid.
    coords : tuple
        Coordinates of point as (x, y)
    mode : {'raise', 'wrap', 'clip'}, optional
        Action to take if a point is outside of the grid.

    Returns
    -------
    array_like :
        Indices of the nodes nearest the given coordinates.

    Examples
    --------
    Create a grid of 4 by 5 nodes with unit spacing.

    >>> import landlab
    >>> from landlab.grid.raster_funcs import find_nearest_node
    >>> rmg = landlab.RasterModelGrid(4, 5)

    The points can be either a tuple of scalars or of arrays.

    >>> find_nearest_node(rmg, (0.2, 0.6))
    5
    >>> find_nearest_node(rmg, (np.array([1.6, 3.6]), np.array([2.3, .7])))
    array([12,  9])

    The *mode* keyword indicates what to do if a point is outside of the
    grid.

    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='raise')
    Traceback (most recent call last):
        ...
    ValueError: invalid entry in coordinates array
    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='clip')
    5
    >>> find_nearest_node(rmg, (-0.6, 0.6), mode='wrap')
    9
    """
    if isinstance(coords[0], np.ndarray):
        return _find_nearest_node_ndarray(rmg, coords, mode=mode)
    else:
        return find_nearest_node(
            rmg, (np.array(coords[0]), np.array(coords[1])), mode=mode)


def _find_nearest_node_ndarray(rmg, coords, mode='raise'):
    """Find the node nearest to a point.

    Parameters
    ----------
    rmg : RasterModelGrid
        A RasterModelGrid.
    coords : tuple of float
        Coordinates of test points as *x*, then *y*.
    mode : {'raise', 'wrap', 'clip'}, optional
        What to do with out-of-bounds indices (as with
        numpy.ravel_multi_index).

    Returns
    -------
    ndarray
        Nodes that are closest to the points.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _find_nearest_node_ndarray
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> grid = RasterModelGrid((4, 5))
    >>> _find_nearest_node_ndarray(grid, (.25, 1.25))
    5
    >>> _find_nearest_node_ndarray(grid, (.75, 2.25))
    11

    >>> grid = RasterModelGrid((4, 5), spacing=(3, 4))
    >>> _find_nearest_node_ndarray(grid, (3.1, 4.1))
    6
    """
    column_indices = np.int_(
        np.around((coords[0] - rmg.node_x[0]) / rmg.dx))
    row_indices = np.int_(
        np.around((coords[1] - rmg.node_y[0]) / rmg.dy))

    return rmg.grid_coords_to_node_id(row_indices, column_indices, mode=mode)


def _value_is_in_bounds(value, bounds):
    """Check if a value is within bounds.

    Parameters
    ----------
    value : float or ndarray
        The test value.
    bounds : (lower, upper)
        The lower and upper bounds.

    Returns
    -------
    bool
        ``True`` if the value is within the bounds. Otherwise, ``False``.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _value_is_in_bounds
    >>> import numpy as np
    >>> _value_is_in_bounds(.5, (0, 1))
    True
    >>> _value_is_in_bounds(1, (0, 1))
    False
    >>> _value_is_in_bounds(0, (0, 1))
    True
    >>> _value_is_in_bounds(np.array((0, 1)), (0, 1))
    array([ True, False], dtype=bool)
    """
    dummy = value >= bounds[0]
    dummy &= value < bounds[1]
    return dummy


def _value_is_within_axis_bounds(rmg, value, axis):
    """Check if a value is within the bounds of a grid axis.

    Parameters
    ----------
    rmg : RasterModelGrid
        A RasterModelGrid.
    value : float
        The test value.
    axis : int
        The axis.

    Returns
    -------
    bool
        ``True`` if the value is within the axis bounds. Otherwise, ``False``.

    Examples
    --------
    >>> from landlab.grid.raster_funcs import _value_is_within_axis_bounds
    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid((4, 5))
    >>> _value_is_within_axis_bounds(rmg, 3.1, 0)
    False
    >>> _value_is_within_axis_bounds(rmg, 2.9, 0)
    True
    >>> _value_is_within_axis_bounds(rmg, 4.1, 1)
    False
    >>> _value_is_within_axis_bounds(rmg, 3.9, 1)
    True
    """
    axis_coord = rmg.node_axis_coordinates(axis)
    return _value_is_in_bounds(value, (axis_coord[0], axis_coord[-1]))


def is_coord_on_grid(rmg, coords, axes=(0, 1)):
    """Check if coordinates are contained on a grid.

    Parameters
    ----------
    rmg : RasterModelGrid
        Source grid.
    coords : tuple
        Coordinates of point as (x, y)
    axes : tuple, optional
        Check bounds only on a particular axis

    Examples
    --------
    Create a grid that ranges from x=0 to x=4, and y=0 to y=3.

    >>> from landlab import RasterModelGrid
    >>> from landlab.grid.raster_funcs import is_coord_on_grid
    >>> grid = RasterModelGrid(4, 5)
    >>> is_coord_on_grid(grid, (3.999, 2.999))
    True

    Check two points with one call. Numpy broadcasting rules apply for the
    point coordinates.

    >>> is_coord_on_grid(grid, ([3.9, 4.1], 2.9))
    array([ True, False], dtype=bool)

    >>> is_coord_on_grid(grid, ([3.9, 4.1], 2.9), axes=(0, ))
    array([ True,  True], dtype=bool)
    """
    coords = np.broadcast_arrays(*coords)

    is_in_bounds = _value_is_within_axis_bounds(rmg, coords[1 - axes[0]],
                                                axes[0])
    for axis in axes[1:]:
        is_in_bounds &= _value_is_within_axis_bounds(rmg, coords[1 - axis],
                                                     axis)

    return is_in_bounds



import numpy as np
from numpy.testing import assert_array_equal
try:
    from nose.tools import assert_is
except ImportError:
    from landlab.testing.tools import assert_is

from landlab import RasterModelGrid


def test_flux_from_south_to_north():
    """Test fluxes that move from south to north."""
    rmg = RasterModelGrid(4, 5)
    active_link_flux = np.array([0., 0., 0.,
                                 0., 0., 0., 0.,
                                 1., 1., 1.,
                                 0., 0., 0., 0.,
                                 3., 3., 3.])
    divs = rmg.calculate_flux_divergence_at_nodes(active_link_flux)

    assert_array_equal(
        divs,
        np.array([0.,  0.,  0.,  0., 0.,
                  0.,  1.,  1.,  1., 0.,
                  0.,  2.,  2.,  2., 0.,
                  0., -3., -3., -3., 0.]))


def test_flux_from_east_to_west():
    """Test fluxes tha move from east to west."""
    rmg = RasterModelGrid(4, 5)
    active_link_flux = np.array([0., 0., 0.,
                                 0., 1., 3., 6., 
                                 0., 0., 0., 
                                 0., 1., 3., 6., 
                                 0., 0., 0.])
    divs = rmg.calculate_flux_divergence_at_nodes(active_link_flux)

    assert_array_equal(
        divs,
        np.array([0., 0., 0., 0.,  0.,
                  0., 1., 2., 3., -6.,
                  0., 1., 2., 3., -6.,
                  0., 0., 0., 0.,  0.]))


def test_out_array():
    """Test out keyword."""
    rmg = RasterModelGrid(4, 5)
    active_link_flux = np.array([0., 0., 0.,
                                 0., 1., 3., 6., 
                                 0., 0., 0., 
                                 0., 1., 3., 6., 
                                 0., 0., 0.])

    divs = np.empty(20)
    rtn_divs = rmg.calculate_flux_divergence_at_nodes(active_link_flux,
                                                      out=divs)

    assert_array_equal(
        divs,
        np.array([0., 0., 0., 0.,  0.,
                  0., 1., 2., 3., -6.,
                  0., 1., 2., 3., -6.,
                  0., 0., 0., 0.,  0.]))
    assert_is(rtn_divs, divs)



"""General decorators for the landlab library.

General Landlab decorators
++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.utils.decorators.use_file_name_or_kwds
    ~landlab.utils.decorators.use_field_name_or_array
    ~landlab.utils.decorators.make_return_array_immutable
    ~landlab.utils.decorators.deprecated
"""

import os
import warnings
from functools import wraps
import textwrap

import numpy as np
import six

from ..core.model_parameter_loader import load_params


def use_file_name_or_kwds(func):

    """Decorate a method so that it takes a file name or keywords.

    Parameters
    ----------
    func : A function
        A method function that accepts a ModelGrid as a first argument.

    Returns
    -------
    function
        A function that takes an optional second argument, a file, from which
        to read keywords.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.utils.decorators import use_file_name_or_kwds

    >>> class MyClass(object):
    ...     @use_file_name_or_kwds
    ...     def __init__(self, grid, kw=0.):
    ...         self.kw = kw

    >>> grid = RasterModelGrid((4, 5))
    >>> foo = MyClass(grid)
    >>> foo.kw
    0.0

    >>> foo = MyClass(grid, "kw: 1945")
    >>> foo.kw
    1945
    >>> foo = MyClass(grid, "kw: 1945", kw=1973)
    >>> foo.kw
    1973

    >>> mpd = \"\"\"
    ... kw: kw value
    ... 1e6
    ... \"\"\"
    >>> foo = MyClass(grid, mpd)
    >>> foo.kw
    1000000.0
    """

    @wraps(func)
    def _wrapped(self, *args, **kwds):
        from ..grid import ModelGrid

        if not isinstance(args[0], ModelGrid):
            raise ValueError('first argument must be a ModelGrid')

        if len(args) == 2:
            warnings.warn(
                "Passing a file to a component's __init__ method is "
                "deprecated. Instead, pass parameters as keywords.",
                category=DeprecationWarning)

            if os.path.isfile(args[1]):
                with open(args[1], 'r') as fp:
                    params = load_params(fp)
            else:
                params = load_params(args[1])
        else:
            params = {}

        params.update(kwds)

        func(self, args[0], **params)

    return _wrapped


class use_field_name_or_array(object):

    """Decorate a function so that it accepts a field name or array.

    Parameters
    ----------
    func : function
        A function that accepts a grid and array as arguments.
    at_element : str
        The element type that the field is defined on ('node', 'cell',
        etc.)

    Returns
    -------
    func
        A wrapped function that accepts a grid and either a field name or
        a numpy array.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((4, 5), spacing=(1, 2))

    >>> def my_func(grid, vals):
    ...     return grid.area_of_cell * vals
    >>> my_func(grid, np.arange(grid.number_of_cells))
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    Decorate the function so that the second argument can be array-like or
    the name of a field contained withing the grid. The decorator takes a
    single argument that is the name (as a `str`) of the grid element that
    the values are defined on ("node", "cell", etc.).

    >>> from landlab.utils.decorators import use_field_name_or_array
    >>> @use_field_name_or_array('cell')
    ... def my_func(grid, vals):
    ...     return grid.area_of_cell * vals

    The array of values now can be list or anything that can be converted to
    a numpy array.

    >>> my_func(grid, [0, 1, 2, 3, 4, 5])
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values doesn't have to be flat.

    >>> vals = np.array([[0, 1, 2], [3, 4, 5]])
    >>> my_func(grid, vals)
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values can be a field name.

    >>> _ = grid.add_field('cell', 'elevation', [0, 1, 2, 3, 4, 5])
    >>> my_func(grid, 'elevation')
    array([  0.,   2.,   4.,   6.,   8.,  10.])
    """

    def __init__(self, at_element):
        """Initialize the decorator.

        Parameters
        ----------
        at_element : str
            The element type that the field is defined on ('node', 'cell',
            etc.)
        """
        self._at = at_element

    def __call__(self, func):
        """Wrap the function."""
        @wraps(func)
        def _wrapped(grid, vals, *args, **kwds):
            """Convert the second argument to an array."""
            if isinstance(vals, six.string_types):
                vals = grid[self._at][vals]
            else:
                vals = np.asarray(vals).flatten()

            return func(grid, vals, *args, **kwds)
        return _wrapped


def make_return_array_immutable(func):
    """Decorate a function so that its return array is read-only.

    Parameters
    ----------
    func : function
        A function that returns a numpy array.

    Returns
    -------
    func
        A wrapped function that returns a read-only view of an array.
    """
    @wraps(func)
    def _wrapped(self, *args, **kwds):
        """Make the returned array read-only."""
        array = func(self, *args, **kwds)
        immutable_array = array.view()
        immutable_array.flags.writeable = False
        return immutable_array
    return _wrapped


# def deprecated(use, version):
#     """Mark a function as deprecated.
# 
#     Parameters
#     ----------
#     use : str
#         Name of replacement function to use.
#     version : str
#         Version number when function was marked as deprecated.
# 
#     Returns
#     -------
#     func
#         A wrapped function that issues a deprecation warning.
#     """
#     def real_decorator(func):
# 
#         def _wrapped(*args, **kwargs):
#             """Warn that the function is deprecated before calling it."""
#             warnings.warn(
#                 "Call to deprecated function {name}.".format(
#                     name=func.__name__), category=DeprecationWarning)
#             return func(*args, **kwargs)
#         _wrapped.__dict__.update(func.__dict__)
# 
#         return _wrapped
# 
#     return real_decorator

def deprecated(use, version):
    """Mark a function as deprecated.

    Parameters
    ----------
    use : str
        Name of replacement function to use.
    version : str
        Version number when function was marked as deprecated.

    Returns
    -------
    func
        A wrapped function that issues a deprecation warning.
    """
    def real_decorator(func):
        warning_str = """
.. note:: This method is deprecated as of Landlab version {ver}.

    Use :func:`{use}` instead.

""".format(ver=version, use=use)

        doc_lines = (func.__doc__ or "").split(os.linesep)

        for lineno, line in enumerate(doc_lines):
            if len(line.rstrip()) == 0:
                break

        head = doc_lines[:lineno]
        body = doc_lines[lineno:]

        head = textwrap.dedent(os.linesep.join(head))
        body = textwrap.dedent(os.linesep.join(body))

        func.__doc__ = os.linesep.join([head, warning_str, body])

        @wraps(func)
        def _wrapped(*args, **kwargs):
            if func.__name__.startswith('_'):
                pass
            else:
                warnings.warn(
                    message="Call to deprecated function {name}.".format(
                        name=func.__name__), category=DeprecationWarning)
            return func(*args, **kwargs)
        _wrapped.__dict__.update(func.__dict__)

        return _wrapped

    return real_decorator
